@@ -25,12 +25,20 @@
 #include <stdarg.h>
 #include <string.h>
 #include <unistd.h>
+#include <pthread.h>
 
 /**
  * progress.h - only for progress bar
  */
 #include "progress.h"
 
+void *thread_update_pui(void *arg);
+progress_bar	prog;		/// progress_bar structure defined in progress.h
+unsigned long long copied;
+unsigned long long block_id;
+int done;
+
+
 /**
  * partclone.h - include some structures like image_head, opt_cmd, ....
  *               and functions for main used.
@@ -39,7 +47,6 @@
 
 /// global variable
 cmd_opt		opt;			/// cmd_opt structure defined in partclone.h
-p_dialog_mesg	m_dialog;			/// dialog format string
 
 /**
  * Include different filesystem header depend on what flag you want.
@@ -99,7 +106,7 @@ int main(int argc, char **argv){
     char*		buffer2;			/// buffer data for malloc used
     int			dfr, dfw;		/// file descriptor for source and target
     int			r_size, w_size;		/// read and write size
-    unsigned long long	block_id, copied = 0;	/// block_id is every block in partition
+    //unsigned long long	block_id, copied = 0;	/// block_id is every block in partition
     /// copied is copied block count
     off_t		offset = 0, sf = 0;	/// seek postition, lseek result
     int			start, stop;		/// start, range, stop number for progress bar
@@ -109,32 +116,37 @@ int main(int argc, char **argv){
     char		bitmagic[8] = "BiTmAgIc";// only for check postition
     char		bitmagic_r[8]="00000000";/// read magic string from image
     int			cmp;			/// compare magic string
-    char		*bitmap;		/// the point for bitmap data
+    unsigned long	*bitmap;		/// the point for bitmap data
     int			debug = 0;		/// debug or not
     unsigned long	crc = 0xffffffffL;	/// CRC32 check code for writint to image
     unsigned long	crc_ck = 0xffffffffL;	/// CRC32 check code for checking
     unsigned long	crc_ck2 = 0xffffffffL;	/// CRC32 check code for checking
     int			c_size;			/// CRC32 code size
     int			n_crc_size = CRC_SIZE;
-    char*		crc_buffer;		/// buffer data for malloc crc code
-    int			done = 0;
+    //int			done = 0;
     int			s_count = 0;
     int			rescue_num = 0;
     unsigned long long			rescue_pos = 0;
     unsigned long long			main_pos = 0;
     int			tui = 0;		/// text user interface
     int			pui = 0;		/// progress mode(default text)
-    int                 next=1,next_int=1,next_max_count=7,next_count=7,next_block=0;
+    int                 next=1,next_int=1,next_max_count=7,next_count=7,i;
     unsigned long long  next_block_id;
     char*               cache_buffer;
     int                 nx_current=0;
+    char                bbuffer[4096];
+    int flag;
+    int pres;
+    pthread_t prog_thread;
+    void *p_result;
 
-    char *bad_sectors_warning_msg =
+    static const char *const bad_sectors_warning_msg =
         "*************************************************************************\n"
-        "* WARNING: The disk has bad sector. This means physical damage on the   *\n"
-        "* disk surface caused by deterioration, manufacturing faults or other   *\n"
-        "* reason. The reliability of the disk may stay stable or degrade fast.  *\n"
-        "* Use the --rescue option to efficiently save as much data as possible! *\n"
+        "* WARNING: The disk has bad sectors. This means physical damage on the  *\n"
+        "* disk surface caused by deterioration, manufacturing faults, or        *\n"
+        "* another reason. The reliability of the disk may remain stable or      *\n"
+        "* degrade quickly. Use the --rescue option to efficiently save as much  *\n"
+        "* data as possible!                                                     *\n"
         "*************************************************************************\n";
 
     image_head		image_hdr;		/// image_head structure defined in partclone.h
@@ -167,9 +179,6 @@ int main(int argc, char **argv){
     if (opt.ncurses){
         pui = NCURSES;
         log_mesg(1, 0, 0, debug, "Using Ncurses User Interface mode.\n");
-    } else if (opt.dialog){
-        pui = DIALOG;
-        log_mesg(1, 0, 0, debug, "Using Dialog User Interface mode.\n");
     } else
         pui = TEXT;
 
@@ -177,8 +186,6 @@ int main(int argc, char **argv){
     if ((opt.ncurses) && (tui == 0)){
         opt.ncurses = 0;
         log_mesg(1, 0, 0, debug, "Open Ncurses User Interface Error.\n");
-    } else if ((opt.dialog) && (tui == 1)){
-        m_dialog.percent = 1;
     }
 
     /// print partclone info
@@ -191,7 +198,7 @@ int main(int argc, char **argv){
 
     /// ignore crc check
     if(opt.ignore_crc)
-	log_mesg(1, 0, 1, debug, "Ignore CRC error\n");
+	log_mesg(1, 0, 1, debug, "Ignore CRC errors\n");
 
     /**
      * open source and target 
@@ -206,12 +213,12 @@ int main(int argc, char **argv){
     target = opt.target;
     dfr = open_source(source, &opt);
     if (dfr == -1) {
-        log_mesg(0, 1, 1, debug, "Erro EXIT.\n");
+        log_mesg(0, 1, 1, debug, "Error exit\n");
     }
 
     dfw = open_target(target, &opt);
     if (dfw == -1) {
-        log_mesg(0, 1, 1, debug, "Error Exit.\n");
+        log_mesg(0, 1, 1, debug, "Error exit\n");
     }
 
     /**
@@ -228,22 +235,21 @@ int main(int argc, char **argv){
 
         /// check memory size
         if (check_mem_size(image_hdr, opt, &needed_mem) == -1)
-            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %i bytes memory\n", needed_mem);
+            log_mesg(0, 1, 1, debug, "There is not enough free memory, partclone suggests you should have %lld bytes memory\n", needed_mem);
 
 	strncpy(image_hdr.version, IMAGE_VERSION, VERSION_SIZE);
 
         /// alloc a memory to restore bitmap
-        bitmap = (char*)malloc(sizeof(char)*image_hdr.totalblock);
+        bitmap = (unsigned long*)calloc(sizeof(unsigned long), LONGS(image_hdr.totalblock));
         if(bitmap == NULL){
             log_mesg(0, 1, 1, debug, "%s, %i, ERROR:%s", __func__, __LINE__, strerror(errno));
         }
-	memset(bitmap, 0, sizeof(char)*image_hdr.totalblock);
 
         log_mesg(2, 0, 0, debug, "initial main bitmap pointer %i\n", bitmap);
         log_mesg(1, 0, 0, debug, "Initial image hdr - read bitmap table\n");
 
         /// read and check bitmap from partition
-        log_mesg(0, 0, 1, debug, "Calculating bitmap...\n");
+        log_mesg(0, 0, 1, debug, "Calculating bitmap... Please wait... ");
         readbitmap(source, image_hdr, bitmap, pui);
 
         needed_size = (unsigned long long)(((image_hdr.block_size+sizeof(unsigned long))*image_hdr.usedblocks)+sizeof(image_hdr)+sizeof(char)*image_hdr.totalblock);
@@ -252,17 +258,26 @@ int main(int argc, char **argv){
 
         log_mesg(2, 0, 0, debug, "check main bitmap pointer %i\n", bitmap);
 
-        log_mesg(1, 0, 1, debug, "Writing super block and bitmap...");
+        log_mesg(1, 0, 0, debug, "Writing super block and bitmap... ");
         // write image_head to image file
         w_size = write_all(&dfw, (char *)&image_hdr, sizeof(image_head), &opt);
         if(w_size == -1)
             log_mesg(0, 1, 1, debug, "write image_hdr to image error\n");
 
         // write bitmap information to image file
-        w_size = write_all(&dfw, bitmap, sizeof(char)*image_hdr.totalblock, &opt);
-        if(w_size == -1)
-            log_mesg(0, 1, 1, debug, "write bitmap to image error\n");
-        log_mesg(1, 0, 1, debug, "done!\n");
+        for (i = 0; i < image_hdr.totalblock; i++){
+            if (pc_test_bit(i, bitmap)){
+                bbuffer[i % sizeof(bbuffer)] = 1;
+            } else {
+                bbuffer[i % sizeof(bbuffer)] = 0;
+            }
+            if (i % sizeof(bbuffer) == sizeof(bbuffer) - 1 || i == image_hdr.totalblock - 1) {
+                w_size = write_all(&dfw, bbuffer, 1 + (i % sizeof(bbuffer)), &opt);
+                if(w_size == -1)
+                    log_mesg(0, 1, 1, debug, "write bitmap to image error\n");
+            }
+        }
+        log_mesg(0, 0, 1, debug, "done!\n");
     } else if (opt.restore){
 
         log_mesg(1, 0, 0, debug, "restore image hdr - get image_head from image file\n");
@@ -272,14 +287,13 @@ int main(int argc, char **argv){
 
         /// check memory size
         if (check_mem_size(image_hdr, opt, &needed_mem) == -1)
-            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %i bytes memory\n", needed_mem);
+            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %lld bytes memory\n", needed_mem);
 
         /// alloc a memory to restore bitmap
-        bitmap = (char*)malloc(sizeof(char)*image_hdr.totalblock);
+        bitmap = (unsigned long*)calloc(sizeof(unsigned long), LONGS(image_hdr.totalblock));
         if(bitmap == NULL){
             log_mesg(0, 1, 1, debug, "%s, %i, ERROR:%s", __func__, __LINE__, strerror(errno));
         }
-	memset(bitmap, 0, sizeof(char)*image_hdr.totalblock);
 
         /// check the image magic
         if (memcmp(image_hdr.magic, IMAGE_MAGIC, IMAGE_MAGIC_SIZE) != 0)
@@ -313,16 +327,15 @@ int main(int argc, char **argv){
 
         /// check memory size
         if (check_mem_size(image_hdr, opt, &needed_mem) == -1)
-            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %i bytes memory\n", needed_mem);
+            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %lld bytes memory\n", needed_mem);
 
 	strncpy(image_hdr.version, IMAGE_VERSION, VERSION_SIZE);
 
         /// alloc a memory to restore bitmap
-        bitmap = (char*)malloc(sizeof(char)*image_hdr.totalblock);
+        bitmap = (unsigned long*)calloc(sizeof(unsigned long), LONGS(image_hdr.totalblock));
         if(bitmap == NULL){
             log_mesg(0, 1, 1, debug, "%s, %i, ERROR:%s", __func__, __LINE__, strerror(errno));
         }
-	memset(bitmap, 0, sizeof(char)*image_hdr.totalblock);
 
         log_mesg(2, 0, 0, debug, "initial main bitmap pointer %i\n", bitmap);
         log_mesg(1, 0, 0, debug, "Initial image hdr - read bitmap table\n");
@@ -337,7 +350,7 @@ int main(int argc, char **argv){
         }
 
         log_mesg(2, 0, 0, debug, "check main bitmap pointer %i\n", bitmap);
-        log_mesg(1, 0, 1, debug, "done!\n");
+        log_mesg(0, 0, 1, debug, "done!\n");
     } else if (opt.domain) {
         log_mesg(1, 0, 0, debug, "Initial image hdr - get Super Block from partition\n");
         log_mesg(1, 0, 1, debug, "Reading Super Block\n");
@@ -347,16 +360,15 @@ int main(int argc, char **argv){
 
         /// check memory size
         if (check_mem_size(image_hdr, opt, &needed_mem) == -1)
-            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %i bytes memory\n", needed_mem);
+            log_mesg(0, 1, 1, debug, "Ther is no enough free memory, partclone suggests you should have %lld bytes memory\n", needed_mem);
 
 	strncpy(image_hdr.version, IMAGE_VERSION, VERSION_SIZE);
 
         /// alloc a memory to restore bitmap
-        bitmap = (char*)malloc(sizeof(char)*image_hdr.totalblock);
+        bitmap = (unsigned long*)calloc(sizeof(unsigned long), LONGS(image_hdr.totalblock));
         if(bitmap == NULL){
             log_mesg(0, 1, 1, debug, "%s, %i, ERROR:%s", __func__, __LINE__, strerror(errno));
         }
-	memset(bitmap, 0, sizeof(char)*image_hdr.totalblock);
 
         log_mesg(2, 0, 0, debug, "initial main bitmap pointer %i\n", bitmap);
         log_mesg(1, 0, 0, debug, "Initial image hdr - read bitmap table\n");
@@ -366,7 +378,7 @@ int main(int argc, char **argv){
         readbitmap(source, image_hdr, bitmap, pui);
 
         log_mesg(2, 0, 0, debug, "check main bitmap pointer %i\n", bitmap);
-        log_mesg(1, 0, 1, debug, "done!\n");
+        log_mesg(0, 0, 1, debug, "done!\n");
     }
 
     log_mesg(1, 0, 0, debug, "print image_head\n");
@@ -381,14 +393,23 @@ int main(int argc, char **argv){
     /**
      * initial progress bar
      */
-    progress_bar	prog;		/// progress_bar structure defined in progress.h
+    //progress_bar	prog;		/// progress_bar structure defined in progress.h
     start = 0;				/// start number of progress bar
-    stop = (image_hdr.usedblocks+1);	/// get the end of progress number, only used block
+    stop = (image_hdr.usedblocks);	/// get the end of progress number, only used block
     log_mesg(1, 0, 0, debug, "Initial Progress bar\n");
     /// Initial progress bar
-    progress_init(&prog, start, stop, image_hdr.block_size);
+    if (opt.no_block_detail)
+	flag = NO_BLOCK_DETAIL;
+    else
+	flag = IO;
+    progress_init(&prog, start, stop, image_hdr.totalblock, flag, image_hdr.block_size);
     copied = 0;				/// initial number is 0
 
+    /**
+     * thread to print progress
+     */
+    pres = pthread_create(&prog_thread, NULL, thread_update_pui, NULL);
+
     /**
      * start read and write data between device and image file
      */
@@ -420,10 +441,10 @@ int main(int argc, char **argv){
 	    main_pos = lseek(dfr, 0, SEEK_CUR);
 	    log_mesg(3, 0, 0, debug, "man pos = %lli\n", main_pos);
 
-            if (bitmap[block_id] == 1){
+            if (pc_test_bit(block_id, bitmap)){
                 /// if the block is used
                 log_mesg(1, 0, 0, debug, "block_id=%lli, ",block_id);
-                log_mesg(2, 0, 0, debug, "bitmap=%i, ",bitmap[block_id]);
+                log_mesg(2, 0, 0, debug, "bitmap=%i, ",pc_test_bit(block_id, bitmap));
 
                 offset = (off_t)(block_id * image_hdr.block_size);
 #ifdef _FILE_OFFSET_BITS
@@ -459,17 +480,11 @@ int main(int argc, char **argv){
                     log_mesg(0, 1, 1, debug, "write error %i \n", w_size);
 
                 /// generate crc32 code and write it.
-                crc_buffer = (char*)malloc(CRC_SIZE); ///alloc a memory to copy data
-                if(crc_buffer == NULL){
-                    log_mesg(0, 1, 1, debug, "%s, %i, ERROR:%s", __func__, __LINE__, strerror(errno));
-                }
                 crc = crc32(crc, buffer, w_size);
+                char crc_buffer[CRC_SIZE];
                 memcpy(crc_buffer, &crc, CRC_SIZE);
                 c_size = write_all(&dfw, crc_buffer, CRC_SIZE, &opt);
 
-                /// free buffer
-                free(crc_buffer);
-
                 copied++;					/// count copied block
                 total_write += (unsigned long long)(w_size);	/// count copied size
                 log_mesg(3, 0, 0, debug, "total=%lli, ", total_write);
@@ -490,14 +505,8 @@ int main(int argc, char **argv){
                 log_mesg(2, 0, 0, debug, "end\n");
 #endif
             }
-	    if (!opt.quiet)
-		update_pui(&prog, copied, done);
         } /// end of for    
-	done = 1;
-	update_pui(&prog, copied, done);
-        sync_data(dfw, &opt);	
-        free(buffer);
-
+	free(buffer);
     } else if (opt.restore) {
 
         /**
@@ -511,7 +520,7 @@ int main(int argc, char **argv){
 
         /// seek to the first
         sf = lseek(dfw, 0, SEEK_SET);
-        log_mesg(1, 0, 0, debug, "seek %lli for writing dtat string\n",sf);
+        log_mesg(1, 0, 0, debug, "seek %lli for writing data string\n",sf);
         if (sf == (off_t)-1)
             log_mesg(0, 1, 1, debug, "seek set %lli\n", sf);
 
@@ -527,16 +536,17 @@ int main(int argc, char **argv){
 
         /// start restore image file to partition
         log_mesg(1, 0, 0, debug, "start restore data...\n");
+
         for( block_id = 0; block_id < image_hdr.totalblock; block_id++ ){
 
             r_size = 0;
             w_size = 0;
 
 
-            if (bitmap[block_id] == 1){ 
+            if (pc_test_bit(block_id, bitmap)){ 
                 /// The block is used
                 log_mesg(1, 0, 0, debug, "block_id=%lli, ",block_id);
-                log_mesg(2, 0, 0, debug, "bitmap=%i, ",bitmap[block_id]);
+                log_mesg(2, 0, 0, debug, "bitmap=%i, ",pc_test_bit(block_id, bitmap));
 
                 memset(buffer, 0, image_hdr.block_size);
                 r_size = read_all(&dfr, buffer, image_hdr.block_size, &opt);
@@ -546,17 +556,13 @@ int main(int argc, char **argv){
 
                 /// read crc32 code and check it.
                 crc_ck = crc32(crc_ck, buffer, r_size);
-                crc_buffer = (char*)malloc(CRC_SIZE); ///alloc a memory to copy data
-                if(crc_buffer == NULL){
-                    log_mesg(0, 1, 1, debug, "%s, %i, ERROR:%s", __func__, __LINE__, strerror(errno));
-                }
+                char crc_buffer[CRC_SIZE];
                 c_size = read_all(&dfr, crc_buffer, CRC_SIZE, &opt);
                 if (c_size < CRC_SIZE)
                     log_mesg(0, 1, 1, debug, "read CRC error: %s, please check your image file. \n", strerror(errno));
 
-                memcpy(&crc, crc_buffer, CRC_SIZE);
 		/*FIX: 64bit image can't ignore crc error*/
-                if ((memcmp(&crc, &crc_ck, CRC_SIZE) != 0) && (!opt.ignore_crc)){
+                if ((memcmp(crc_buffer, &crc_ck, CRC_SIZE) != 0) && (!opt.ignore_crc)){
                     log_mesg(1, 0, 0, debug, "CRC Check error. 64bit bug before v0.1.0 (Rev:250M), enlarge crc size and recheck again....\n ");
                     /// check again
                     buffer2 = (char*)malloc(image_hdr.block_size+CRC_SIZE); ///alloc a memory to copy data
@@ -571,8 +577,7 @@ int main(int argc, char **argv){
                     c_size = read_all(&dfr, crc_buffer, CRC_SIZE, &opt);
                     if (c_size < CRC_SIZE)
                         log_mesg(0, 1, 1, debug, "read CRC error: %s, please check your image file. \n", strerror(errno));
-                    memcpy(&crc, crc_buffer, CRC_SIZE);
-                    if ((memcmp(&crc, &crc_ck2, CRC_SIZE) != 0 )&& (!opt.ignore_crc)) {
+                    if ((memcmp(crc_buffer, &crc_ck2, CRC_SIZE) != 0 )&& (!opt.ignore_crc)) {
                         log_mesg(0, 1, 1, debug, "CRC error again at %i...\n ", sf);
                     } else {
                         crc_ck = crc_ck2;
@@ -587,10 +592,9 @@ int main(int argc, char **argv){
 		    for (next_int = 1; next_int <= next_max_count; next_int++)
 		    {
 			next_block_id = block_id+next_int;
-			next_block = bitmap[next_block_id];
-			if (next_block == 1) {
+			if (pc_test_bit(next_block_id, bitmap)) {
 			    next++;
-			} else if (next_block == 0){
+			} else {
 			    next_count = next;
 			    break;
 			}
@@ -622,8 +626,6 @@ int main(int argc, char **argv){
 		    nx_current=0;
 		}
 
-                /// free buffer
-                free(crc_buffer);
                 copied++;					/// count copied block
                 total_write += (unsigned long long) w_size;	/// count copied size
 
@@ -648,14 +650,8 @@ int main(int argc, char **argv){
 #endif
 		}
             }
-	    if (!opt.quiet)
-		update_pui(&prog, copied, done);
         } // end of for
-	/// free buffer
 	free(buffer);
-	done = 1;
-	update_pui(&prog, copied, done);
-        sync_data(dfw, &opt);	
     } else if (opt.dd){
         sf = lseek(dfr, 0, SEEK_SET);
         log_mesg(1, 0, 0, debug, "seek %lli for reading data string\n",sf);
@@ -678,11 +674,11 @@ int main(int argc, char **argv){
             r_size = 0;
             w_size = 0;
 
-            if (bitmap[block_id] == 1){
+            if (pc_test_bit(block_id, bitmap)){
                 /// if the block is used
 
                 log_mesg(1, 0, 0, debug, "block_id=%lli, ",block_id);
-                log_mesg(2, 0, 0, debug, "bitmap=%i, ",bitmap[block_id]);
+                log_mesg(2, 0, 0, debug, "bitmap=%i, ",pc_test_bit(block_id, bitmap));
                 offset = (off_t)(block_id * image_hdr.block_size);
 #ifdef _FILE_OFFSET_BITS
                 sf = lseek(dfr, offset, SEEK_SET);
@@ -737,14 +733,8 @@ int main(int argc, char **argv){
                     log_mesg(0, 1, 1, debug, "clone seek error %lli errno=%i\n", (long long)offset, (int)errno);
 #endif
             }
-	    if (!opt.quiet)
-		update_pui(&prog, copied, done);
         } /// end of for
-	done = 1;
-	update_pui(&prog, copied, done);
-	/// free buffer
 	free(buffer);
-        sync_data(dfw, &opt);
     } else if (opt.domain) {
         log_mesg(0, 0, 0, debug, "Total block %i\n", image_hdr.totalblock);
         log_mesg(1, 0, 0, debug, "start writing domain log...\n");
@@ -758,10 +748,10 @@ int main(int argc, char **argv){
         dprintf(dfw, "#      pos        size  status\n");
         // start logging the used/unused areas
         next_block_id = 0;
-        cmp = bitmap[0];
+        cmp = pc_test_bit(0, bitmap);
         for( block_id = 0; block_id <= image_hdr.totalblock; block_id++ ){
             if (block_id < image_hdr.totalblock){
-                nx_current = bitmap[block_id];
+                nx_current = pc_test_bit(block_id, bitmap);
                 if (nx_current == 1)
                     copied++;
             } else
@@ -776,11 +766,12 @@ int main(int argc, char **argv){
             }
             // don't bother updating progress
         } /// end of for
-        done = 1;
-        update_pui(&prog, copied, done);
-        sync_data(dfw, &opt);
     }
 
+    done = 1;
+    pres = pthread_join(prog_thread, &p_result);
+    update_pui(&prog, copied, block_id, done);
+    sync_data(dfw, &opt);	
     print_finish_info(opt);
 
     close (dfr);    /// close source
@@ -795,3 +786,12 @@ int main(int argc, char **argv){
 #endif
     return 0;	    /// finish
 }
+
+void *thread_update_pui(void *arg){
+
+    while (done == 0) {
+        if(!opt.quiet)
+		update_pui(&prog, copied, block_id, done);
+    }
+    pthread_exit("exit");
+}
